import stan
import pandas as pd
import numpy as np
import scipy.stats as ss
import matplotlib.pyplot as plt
import seaborn as sb
import arviz as az


def calcMAPEst(samples, a = 0.05, n_points = 1000):
    """　MAP推定値とMAP推定値における推定確率密度関数の値を求める
         samples:    MAP推定値を求めるサンプル
         a:　　　　　サンプルの左右a/2を除く。デフォルト値=0.05
         n_points:   KDEによる確率密度関数値を求める点の数。
         　　　　　　デフォルト値=1000
    """
    import numpy as np
    import scipy.stats as ss

    Lp, Up = np.percentile(samples, [100 * a/2, 100 * (1 - a/2)])
    coord = np.linspace(Lp, Up, n_points)
    est_pdf = ss.gaussian_kde(samples).pdf(coord) 
    map_idx = np.argmax(est_pdf)
    MAP_Est = coord[map_idx]      
    return MAP_Est, est_pdf[map_idx]    #   MAP推定値とMAP推定値における確率密度関数値


finnm = input('Input Data File (*.xlsx) = ')
xlsx_fl = pd.ExcelFile(finnm)
data_xlsx = pd.read_excel(xlsx_fl,  header = None)
data = data_xlsx.values

foutnm = input('Output Text File (*.txt) = ')
fout = open(foutnm, 'w')
fout.write('Input Data File = {}\n'.format(finnm))

print('data[0] =\n', data[0])
St = data[1:,1]
Res = data[1:,2]
fout.write('\nData...\n')
for v in zip(St,Res):
    print(v)
    fout.write(f'{v[0]:>10}{v[1]:>10}\n')

N = len(St)

def f_init():
    p_mu = np.median(St)
    p_sgm = np.std(St)
    p_C1 = p_mu * 0.9   
    d_v = p_mu * 0.1
    return {'v_sgm':p_sgm/100.0,
            'v_C1':p_C1/100.0, 'v_mu': d_v/100.0,
            'v_C2':d_v/100.0}  

Data = {'N':N, 'St':St, 'Res':Res}

with open('pm4cat.stan', 'r') as fstan:
    sm = stan.build(fstan.read(), data = Data)
fit = sm.sample(num_samples = 5000, num_chains = 4,
                init = [f_init()]*4)

i_data = az.from_pystan(posterior = fit, posterior_model = sm)
smry = az.summary(i_data, var_names = ['C1', 'mu', 'C2'])
print('Summary =\n',smry)
fout.write(f'\nSummary:\n{smry}\n')

az.plot_trace(i_data, var_names = ['C1', 'mu', 'C2', 'sgm'])  
plt.show()

d_fm = fit.to_frame()

map_C1 = calcMAPEst(d_fm['C1'])[0]
map_mu = calcMAPEst(d_fm['mu'])[0]
map_C2 = calcMAPEst(d_fm['C2'])[0]
sb.kdeplot(d_fm['C1'], label = f'C1(MAP={map_C1:.2f})')
sb.kdeplot(d_fm['mu'], label = f'$\mu$(MAP={map_mu:.2f})')
sb.kdeplot(d_fm['C2'], label = f'C2(MAP={map_C2:.2f})')
plt.xlabel('')
plt.legend()
plt.show()

diffC2muC1 = d_fm['C2'] + d_fm['C1'] - (d_fm['mu'] * 2)
prop = np.mean([1 if v > 0.0 else 0 for v in diffC2muC1])
map_diff = calcMAPEst(diffC2muC1)[0]
sb.kdeplot(diffC2muC1)
plt.title(f'(C2-mu)-(mu-C1): MAP={map_diff:.2f}' +
          f'\nP(>0) = {prop:.3f}', fontsize = 16)
plt.show()

map_sgm = calcMAPEst(d_fm['sgm'])[0]
sb.kdeplot(d_fm['sgm'], label = f'$\sigma$(MAP={map_sgm:.2f})')
plt.xlabel('')
plt.legend()
plt.show()

min_St = np.min(St)   
max_St = np.max(St) 
min_x = min_St - 0.05 * (max_St - min_St)
max_x = max_St + 0.05 * (max_St - min_St)

plt.axis([min_x, max_x, 0.0, 1.0])

phi = np.linspace(min_x, max_x, 1000)
psi1 = ss.norm.cdf(phi, map_C1, map_sgm)
psiMu = ss.norm.cdf(phi, map_mu, map_sgm)
psi2 = ss.norm.cdf(phi, map_C2, map_sgm)
p_st = []
p_res = []
for i in range(0, N):
    p_st.append(St[i])
    if Res[i] == 1:
        p_res.append(0.1 * np.random.uniform())
    elif Res[i] == 2:
        p_res.append(1.0 / 3.0 - 0.05 + 0.1 * np.random.uniform())
    elif Res[i] == 3:
        p_res.append(2.0 / 3.0 - 0.05 + 0.1 * np.random.uniform())            
    else:
        p_res.append(1.0 - 0.1 * np.random.uniform())
                
        
plt.title(("Psychometric functions   JND = {0:.5}" +
           "\nC1 = {1:.5}   PSE = {2:.5}   C2 = {3:.5}").
          format(map_sgm * 0.67449, map_C1, map_mu, map_C2)) 
          
plt.plot(phi, psi1, 'b--', linewidth = 5)
plt.plot(phi, psiMu, 'g', linewidth = 5)
plt.plot(phi, psi2, 'b-.', linewidth = 5)
plt.plot(p_st, p_res, 'go')
plt.xlabel('$S_{comp}$', fontsize = 14)
plt.legend(['PM1', 'PM', 'PM2', 'Data'], loc = 2)
plt.show()

fout.close()
print()
print(foutnm + ' was saved.\n')

